import { __spreadArray, __read, __assign } from 'tslib';
import sync, { cancelSync, flushSync } from 'framesync';
import { mix } from 'popmotion';
import { animate } from '../../animation/animate.js';
import { SubscriptionManager } from '../../utils/subscription-manager.js';
import { mixValues } from '../animation/mix-values.js';
import { copyBoxInto } from '../geometry/copy.js';
import { applyBoxDelta, applyTreeDeltas } from '../geometry/delta-apply.js';
import { calcBoxDelta, calcLength } from '../geometry/delta-calc.js';
import { removeBoxTransforms } from '../geometry/delta-remove.js';
import { createDelta, createBox } from '../geometry/models.js';
import { translateAxis, transformBox } from '../geometry/operations.js';
import { getValueTransition } from '../../animation/utils/transitions.js';
import { isDeltaZero } from '../geometry/utils.js';
import { NodeStack } from '../shared/stack.js';
import { scaleCorrectors } from '../styles/scale-correction.js';
import { buildProjectionTransform } from '../styles/transform.js';
import { eachAxis } from '../utils/each-axis.js';
import { hasTransform, hasScale } from '../utils/has-transform.js';

function createProjectionNode(_a) {
    var attachResizeListener = _a.attachResizeListener, defaultParent = _a.defaultParent, measureScroll = _a.measureScroll, resetTransform = _a.resetTransform;
    return /** @class */ (function () {
        function ProjectionNode(id, latestValues, parent) {
            var _this = this;
            if (parent === void 0) { parent = defaultParent === null || defaultParent === void 0 ? void 0 : defaultParent(); }
            this.children = new Set();
            this.options = {};
            this.isLayoutDirty = false;
            this.isUpdating = false;
            this.shouldResetTransform = false;
            this.treeScale = { x: 1, y: 1 }; // TODO Lazy-initialise
            // Note: Currently only running on root node
            this.potentialNodes = new Map();
            this.updateProjection = function () {
                updateProjectionTree(_this);
            };
            this.isVisible = true;
            /**
             * Animation
             */
            this.animationProgress = 0;
            /**
             * Shared layout
             */
            // TODO Only running on root node
            this.sharedNodes = new Map();
            this.id = id;
            this.latestValues = latestValues;
            this.root = parent ? parent.root || parent : this;
            this.path = parent ? __spreadArray(__spreadArray([], __read(parent.path)), [parent]) : [];
            this.parent = parent;
            this.root.registerPotentialNode(id, this);
        }
        ProjectionNode.prototype.registerPotentialNode = function (id, node) {
            this.potentialNodes.set(id, node);
        };
        /**
         * Lifecycles
         */
        ProjectionNode.prototype.mount = function (instance, isLayoutDirty) {
            var _this = this;
            var _a;
            if (isLayoutDirty === void 0) { isLayoutDirty = false; }
            if (this.instance)
                return;
            this.instance = instance;
            var _b = this.options, layoutId = _b.layoutId, layout = _b.layout, visualElement = _b.visualElement;
            if (visualElement && !visualElement.getInstance()) {
                visualElement.mount(instance);
            }
            (_a = this.parent) === null || _a === void 0 ? void 0 : _a.children.add(this);
            this.root.potentialNodes.delete(this.id);
            if (isLayoutDirty) {
                this.isLayoutDirty = true;
                this.setTargetDelta(createDelta());
            }
            attachResizeListener === null || attachResizeListener === void 0 ? void 0 : attachResizeListener(instance, function () {
                // TODO: Complete all active animations/delete all projections
            });
            if (layoutId) {
                this.root.registerSharedNode(layoutId, this);
            }
            // Only register the handler if it requires layout animation
            if (visualElement && (layoutId || layout)) {
                this.onLayoutDidUpdate(function (_a) {
                    var delta = _a.delta, hasLayoutChanged = _a.hasLayoutChanged, snapshot = _a.snapshot;
                    // TODO: Check here if an animation exists
                    var layoutTransition = visualElement.getDefaultTransition() ||
                        defaultLayoutTransition;
                    var onLayoutAnimationComplete = visualElement.getProps().onLayoutAnimationComplete;
                    if (hasLayoutChanged
                    /**
                     * Don't create a new animation if the target box
                     * hasn't changed TODO: And we're already animating
                     */
                    // !boxEquals(layoutTarget.current, newLayout)
                    ) {
                        // TODO: On final frame, delete delta
                        _this.setAnimationOrigin(delta, snapshot.latestValues);
                        _this.startAnimation(__assign(__assign({}, getValueTransition(layoutTransition, "layout")), { onComplete: onLayoutAnimationComplete }));
                    }
                });
            }
        };
        ProjectionNode.prototype.unmount = function () {
            var _a, _b;
            // @TODO: only snapshot if there's a pending lead sharing the same layout id
            if (this.options.layoutId) {
                this.willUpdate();
            }
            (_a = this.getStack()) === null || _a === void 0 ? void 0 : _a.remove(this);
            (_b = this.parent) === null || _b === void 0 ? void 0 : _b.children.delete(this);
            cancelSync.preRender(this.updateProjection);
        };
        // Note: currently only running on root node
        ProjectionNode.prototype.startUpdate = function () {
            this.isUpdating = true;
            // TODO: Traverse the tree, reset rotations
        };
        ProjectionNode.prototype.willUpdate = function (shouldNotifyListeners) {
            var _a;
            if (shouldNotifyListeners === void 0) { shouldNotifyListeners = true; }
            !this.root.isUpdating && this.root.startUpdate();
            if (!this.isLead() || this.isLayoutDirty)
                return;
            this.isLayoutDirty = true;
            this.path.forEach(function (node) {
                node.shouldResetTransform = true;
                /**
                 * TODO: Check we haven't updated the scroll
                 * since the last didUpdate
                 */
                node.updateScroll();
            });
            this.updateSnapshot();
            shouldNotifyListeners && ((_a = this.layoutWillUpdateListeners) === null || _a === void 0 ? void 0 : _a.notify());
        };
        // Note: Currently only running on root node
        ProjectionNode.prototype.didUpdate = function () {
            if (!this.isUpdating)
                return;
            this.potentialNodes.forEach(function (node, id) {
                var element = document.querySelector("[data-projection-id=\"" + id + "\"]");
                if (element)
                    node.mount(element, true);
                console.log(!!element, id);
            });
            this.potentialNodes.clear();
            /**
             * Write
             */
            resetTreeTransform(this);
            /**
             * Read ==================
             */
            // Update layout measurements of updated children
            updateTreeLayout(this);
            /**
             * Write
             */
            // Notify listeners that the layout is updated
            notifyLayoutUpdate(this);
            // Flush any scheduled updates
            flushSync.update();
            flushSync.preRender();
            flushSync.render();
            this.isUpdating = false;
        };
        ProjectionNode.prototype.scheduleUpdateProjection = function () {
            sync.preRender(this.updateProjection, false, true);
        };
        /**
         * Update measurements
         */
        ProjectionNode.prototype.updateSnapshot = function () {
            if (this.snapshot)
                return;
            var visible = this.removeTransform(this.measure());
            this.snapshot = {
                visible: visible,
                layout: this.removeElementScroll(visible),
                latestValues: {},
            };
        };
        ProjectionNode.prototype.updateLayout = function () {
            // TODO: Incorporate into a forwarded
            // scroll offset
            if (this.options.shouldMeasureScroll) {
                this.updateScroll();
            }
            if (!this.isLayoutDirty)
                return;
            this.layout = this.removeElementScroll(this.measure());
            this.layoutCorrected = createBox();
            this.isLayoutDirty = false;
            this.projectionDelta = undefined;
        };
        ProjectionNode.prototype.updateScroll = function () {
            if (!measureScroll)
                return;
            this.scroll = measureScroll(this.instance);
        };
        ProjectionNode.prototype.resetTransform = function () {
            var _a, _b;
            if (resetTransform &&
                (this.isLayoutDirty || this.shouldResetTransform) &&
                ((this.projectionDelta && !isDeltaZero(this.projectionDelta)) ||
                    hasTransform(this.latestValues))) {
                resetTransform(this.instance);
                this.shouldResetTransform = false;
                (_b = (_a = this.options).onProjectionUpdate) === null || _b === void 0 ? void 0 : _b.call(_a);
            }
        };
        ProjectionNode.prototype.measure = function () {
            var visualElement = this.options.visualElement;
            if (!visualElement)
                return createBox();
            var box = visualElement.measureViewportBox();
            // Remove window scroll to give page-relative coordinates
            var scroll = this.root.scroll;
            // TODO Make loop
            scroll && eachAxis(function (axis) { return translateAxis(box[axis], scroll[axis]); });
            return box;
        };
        ProjectionNode.prototype.removeElementScroll = function (box) {
            var boxWithoutScroll = createBox();
            copyBoxInto(boxWithoutScroll, box);
            /**
             * Performance TODO: Keep a cumulative scroll offset down the tree
             * rather than loop back up the path.
             */
            for (var i = 0; i < this.path.length; i++) {
                var node = this.path[i];
                var scroll_1 = node.scroll, options = node.options;
                if (node !== this.root &&
                    scroll_1 &&
                    options.shouldMeasureScroll) {
                    translateAxis(boxWithoutScroll.x, scroll_1.x);
                    translateAxis(boxWithoutScroll.y, scroll_1.y);
                }
            }
            return boxWithoutScroll;
        };
        ProjectionNode.prototype.removeTransform = function (box) {
            var boxWithoutTransform = createBox();
            copyBoxInto(boxWithoutTransform, box);
            for (var i = 0; i < this.path.length; i++) {
                var node = this.path[i];
                if (!hasTransform(node.latestValues))
                    continue;
                hasScale(node.latestValues) && node.updateSnapshot();
                removeBoxTransforms(boxWithoutTransform, node.latestValues, node.snapshot.layout);
            }
            removeBoxTransforms(boxWithoutTransform, this.latestValues);
            return boxWithoutTransform;
        };
        /**
         *
         */
        ProjectionNode.prototype.setTargetDelta = function (delta) {
            this.targetDelta = delta;
            this.root.scheduleUpdateProjection();
        };
        ProjectionNode.prototype.setOptions = function (options) {
            var _a;
            this.options = __assign(__assign({}, options), { crossfade: (_a = options.crossfade) !== null && _a !== void 0 ? _a : true });
        };
        /**
         * Frame calculations
         */
        ProjectionNode.prototype.resolveTargetDelta = function () {
            if (!this.targetDelta || !this.layout)
                return;
            if (!this.target) {
                this.target = createBox();
                this.targetWithTransforms = createBox();
            }
            copyBoxInto(this.target, this.layout);
            applyBoxDelta(this.target, this.targetDelta);
        };
        ProjectionNode.prototype.calcProjection = function () {
            var target = this.getLead().target;
            if (!this.layout || !target)
                return;
            if (!this.projectionDelta) {
                this.projectionDelta = createDelta();
                this.projectionDeltaWithTransform = createDelta();
            }
            /**
             * Reset the corrected box with the latest values from box, as we're then going
             * to perform mutative operations on it.
             */
            copyBoxInto(this.layoutCorrected, this.layout);
            /**
             * Apply all the parent deltas to this box to produce the corrected box. This
             * is the layout box, as it will appear on screen as a result of the transforms of its parents.
             */
            applyTreeDeltas(this.layoutCorrected, this.treeScale, this.path);
            /**
             * Update the delta between the corrected box and the target box before user-set transforms were applied.
             * This will allow us to calculate the corrected borderRadius and boxShadow to compensate
             * for our layout reprojection, but still allow them to be scaled correctly by the user.
             * It might be that to simplify this we may want to accept that user-set scale is also corrected
             * and we wouldn't have to keep and calc both deltas, OR we could support a user setting
             * to allow people to choose whether these styles are corrected based on just the
             * layout reprojection or the final bounding box.
             */
            calcBoxDelta(this.projectionDelta, this.layoutCorrected, target, this.latestValues);
            // TODO Make this event listener
            var onProjectionUpdate = this.options.onProjectionUpdate;
            onProjectionUpdate && onProjectionUpdate();
        };
        ProjectionNode.prototype.hide = function () {
            this.isVisible = false;
            // TODO: Schedule render
        };
        ProjectionNode.prototype.show = function () {
            this.isVisible = true;
            // TODO: Schedule render
        };
        ProjectionNode.prototype.scheduleRender = function () {
            var _a, _b;
            // TODO Rename this option
            (_b = (_a = this.options).onProjectionUpdate) === null || _b === void 0 ? void 0 : _b.call(_a);
        };
        ProjectionNode.prototype.setAnimationOrigin = function (delta, latestValues) {
            var _this = this;
            var _a;
            this.animationValues = {};
            var targetDelta = createDelta();
            var shouldCrossfadeOpacity = this.options.crossfade === true &&
                !this.path.some(hasOpacityCrossfade);
            var start = (_a = this.snapshot) === null || _a === void 0 ? void 0 : _a.latestValues;
            this.mixTargetDelta = function (latest) {
                var progress = latest / 1000;
                mixAxisDelta(targetDelta.x, delta.x, progress);
                mixAxisDelta(targetDelta.y, delta.y, progress);
                if (latestValues !== _this.latestValues &&
                    _this.animationValues) {
                    mixValues(_this.animationValues, latestValues, _this.latestValues, progress, shouldCrossfadeOpacity, start);
                }
                _this.setTargetDelta(targetDelta);
                _this.root.scheduleUpdateProjection();
            };
            this.mixTargetDelta(0);
        };
        ProjectionNode.prototype.startAnimation = function (options) {
            var _this = this;
            var _a;
            (_a = this.currentAnimation) === null || _a === void 0 ? void 0 : _a.stop();
            this.currentAnimation = animate(0, 1000, __assign(__assign({}, options), { onUpdate: function (latest) {
                    var _a;
                    _this.mixTargetDelta(latest);
                    (_a = options.onUpdate) === null || _a === void 0 ? void 0 : _a.call(options, latest);
                }, onComplete: function () {
                    var _a;
                    _this.animationValues = undefined;
                    (_a = options.onComplete) === null || _a === void 0 ? void 0 : _a.call(options);
                } }));
        };
        ProjectionNode.prototype.applyTransformsToTarget = function () {
            var _a = this.getLead(), targetWithTransforms = _a.targetWithTransforms, target = _a.target, latestValues = _a.latestValues;
            copyBoxInto(targetWithTransforms, target);
            /**
             * Apply the latest user-set transforms to the targetBox to produce the targetBoxFinal.
             * This is the final box that we will then project into by calculating a transform delta and
             * applying it to the corrected box.
             */
            transformBox(targetWithTransforms, latestValues);
            /**
             * Update the delta between the corrected box and the final target box, after
             * user-set transforms are applied to it. This will be used by the renderer to
             * create a transform style that will reproject the element from its actual layout
             * into the desired bounding box.
             */
            calcBoxDelta(this.projectionDeltaWithTransform, this.layoutCorrected, targetWithTransforms, latestValues);
        };
        /**
         * Events
         *
         * TODO Replace this with a key-based lookup
         */
        ProjectionNode.prototype.onLayoutWillUpdate = function (handler) {
            if (!this.layoutWillUpdateListeners) {
                this.layoutWillUpdateListeners = new SubscriptionManager();
            }
            return this.layoutWillUpdateListeners.add(handler);
        };
        ProjectionNode.prototype.onLayoutDidUpdate = function (handler) {
            if (!this.layoutDidUpdateListeners) {
                this.layoutDidUpdateListeners = new SubscriptionManager();
            }
            return this.layoutDidUpdateListeners.add(handler);
        };
        ProjectionNode.prototype.registerSharedNode = function (layoutId, node) {
            if (!this.sharedNodes.has(layoutId)) {
                this.sharedNodes.set(layoutId, new NodeStack());
            }
            var stack = this.sharedNodes.get(layoutId);
            stack.add(node);
        };
        ProjectionNode.prototype.isLead = function () {
            var stack = this.getStack();
            return stack ? stack.lead === this : true;
        };
        ProjectionNode.prototype.getLead = function () {
            var _a;
            var layoutId = this.options.layoutId;
            return layoutId ? ((_a = this.getStack()) === null || _a === void 0 ? void 0 : _a.lead) || this : this;
        };
        ProjectionNode.prototype.getStack = function () {
            var layoutId = this.options.layoutId;
            if (layoutId)
                return this.root.sharedNodes.get(layoutId);
        };
        ProjectionNode.prototype.promote = function () {
            var stack = this.getStack();
            if (stack)
                stack.promote(this);
        };
        ProjectionNode.prototype.getProjectionStyles = function () {
            var _a, _b, _c;
            // TODO: Return lifecycle-persistent object
            var styles = {};
            if (!this.isVisible) {
                return { visibility: "hidden" };
            }
            else {
                styles.visibility = "";
            }
            if (!this.projectionDelta) {
                return emptyStyles;
            }
            // TODO: Return persistent mutable object
            this.applyTransformsToTarget();
            styles.transform = buildProjectionTransform(this.projectionDeltaWithTransform, this.treeScale, this.latestValues);
            var transformTemplate = (_a = this.options.visualElement) === null || _a === void 0 ? void 0 : _a.getProps().transformTemplate;
            if (transformTemplate) {
                styles.transform = transformTemplate(this.latestValues, styles.transform);
            }
            var lead = this.getLead();
            var valuesToRender = lead.animationValues || lead.latestValues;
            // TODO Move into stand-alone, testable function
            var _d = this.projectionDelta, x = _d.x, y = _d.y;
            styles.transformOrigin = x.origin * 100 + "% " + y.origin * 100 + "% 0";
            if (!lead.animationValues) {
                // snap opacity to full 0/1 when the animation is finished
                styles.opacity =
                    lead === this
                        ? (_b = valuesToRender.opacity) !== null && _b !== void 0 ? _b : ""
                        : (_c = valuesToRender.opacityExit) !== null && _c !== void 0 ? _c : 0;
            }
            else {
                styles.opacity =
                    lead === this
                        ? valuesToRender.opacity
                        : valuesToRender.opacityExit;
            }
            /**
             * Apply scale correction
             */
            for (var key in scaleCorrectors) {
                if (valuesToRender[key] === undefined) {
                    continue;
                }
                var _e = scaleCorrectors[key], correct = _e.correct, applyTo = _e.applyTo;
                var corrected = correct(valuesToRender[key], lead);
                if (applyTo) {
                    var num = applyTo.length;
                    for (var i = 0; i < num; i++) {
                        styles[applyTo[i]] = corrected;
                    }
                }
                else {
                    styles[key] = corrected;
                }
            }
            return styles;
        };
        return ProjectionNode;
    }());
}
var emptyStyles = {};
function updateTreeLayout(node) {
    node.updateLayout();
    node.children.forEach(updateTreeLayout);
}
function notifyLayoutUpdate(node) {
    var layout = node.layout, snapshot = node.snapshot;
    if (node.isLead() && layout && snapshot && node.layoutDidUpdateListeners) {
        // TODO Maybe we want to also resize the layout snapshot so we don't trigger
        // animations for instance if layout="size" and an element has only changed position
        if (node.options.animationType === "size") {
            eachAxis(function (axis) {
                var axisSnapshot = snapshot.visible[axis];
                var length = calcLength(axisSnapshot);
                axisSnapshot.min = layout[axis].min;
                axisSnapshot.max = axisSnapshot.min + length;
            });
        }
        else if (node.options.animationType === "position") {
            eachAxis(function (axis) {
                var axisSnapshot = snapshot.visible[axis];
                var length = calcLength(layout[axis]);
                axisSnapshot.max = axisSnapshot.min + length;
            });
        }
        var layoutDelta = createDelta();
        calcBoxDelta(layoutDelta, layout, snapshot.layout);
        var visualDelta = createDelta();
        calcBoxDelta(visualDelta, layout, snapshot.visible);
        node.layoutDidUpdateListeners.notify({
            layout: layout,
            snapshot: snapshot,
            delta: visualDelta,
            hasLayoutChanged: !isDeltaZero(layoutDelta),
        });
    }
    node.children.forEach(notifyLayoutUpdate);
    node.snapshot = undefined;
}
function resetTreeTransform(node) {
    node.resetTransform();
    node.children.forEach(resetTreeTransform);
}
function updateProjectionTree(node) {
    resolveTreeTargetDeltas(node);
    calcTreeProjection(node);
}
function resolveTreeTargetDeltas(node) {
    node.resolveTargetDelta();
    node.children.forEach(resolveTreeTargetDeltas);
}
function calcTreeProjection(node) {
    node.calcProjection();
    node.children.forEach(calcTreeProjection);
}
function mixAxisDelta(output, delta, p) {
    output.translate = mix(delta.translate, 0, p);
    output.scale = mix(delta.scale, 1, p);
    output.origin = delta.origin;
    output.originPoint = delta.originPoint;
}
function hasOpacityCrossfade(node) {
    return (node.animationValues && node.animationValues.opacityExit !== undefined);
}
var defaultLayoutTransition = {
    duration: 0.45,
    ease: [0.4, 0, 0.1, 1],
};

export { createProjectionNode, mixAxisDelta };
